import pprint


class ApiResponse:
    def __init__(self, payload=None, nextToken=None, **kwargs):
        # self.payload = payload or kwargs
        self.payload = payload
        self.headers = kwargs
        self.next_token = nextToken

    def __str__(self):
        return pprint.pformat(self.__dict__)

    def set_next_token(self, nextToken=None):
        if nextToken:
            self.next_token = nextToken
        else:
            self.next_token = self.payload.get("NextToken", None)
        return self.next_token

    @staticmethod
    def set_rate_limit(headers: dict = None):
        try:
            return headers["x-amzn-RateLimit-Limit"]
        except (AttributeError, KeyError, TypeError):
            return None



from python_paypal_api.base import (
    Client,
    PaypalEndpoint,
    PaypalEndpointParams,
    ApiResponse,
    Utils,
    ProductType,
    CategoryType
)

# import logging
# import json


PRODUCT_PROPERTIES = { 
'name', 'type', 'id', 'category', 'home_url',
'image_url', 'description', 'update_time', 'create_time'
}

# PRODUCT_UPDATE = {'category', 'home_url', 'image_url', 'description'}
PRODUCT_UPDATE = {'name', 'category', 'home_url', 'image_url', 'description'}


class Catalog(Client):

    @PaypalEndpoint('/v1/catalogs/products', method='GET')
    def list_products(self, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        return self._request(kwargs.pop('path'), params=kwargs, headers=headers)

    @PaypalEndpoint('/v1/catalogs/products/{}', method='GET')
    def get_product(self, productId:str, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(productId), params=kwargs, headers=headers)

    @PaypalEndpoint('/v1/catalogs/products/{}', method='PATCH')
    def update_product(self, productId:str, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(productId), data=Utils.convert_body(kwargs.pop('body'), False), params=kwargs, headers=headers)

    @PaypalEndpoint('/v1/catalogs/products/{}', method='PATCH')
    def update_product_helper(self, productId:str, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(productId), data=Utils.convert_updates(kwargs.pop('body')), params=kwargs, headers=headers)

    @PaypalEndpoint('/v1/catalogs/products', method='POST')
    def post_product(self, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        Utils.payPalRequestId(headers)
        prefer = "return=representation" # "return=minimal"
        Utils.prefer(headers, prefer)
        return self._request(kwargs.pop('path'), data=Utils.convert_body(kwargs.pop('body'), False), params=kwargs, headers=headers)

    @PaypalEndpoint('/v1/catalogs/products', method='POST')
    def post_product_helper(self, body:(str or dict or list), **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        Utils.payPalRequestId(headers)
        prefer = "return=representation" # "return=minimal"
        Utils.prefer(headers, prefer)
        dictionary = body.copy()

        if isinstance(dictionary["type"], ProductType):
           dictionary["type"] = dictionary["type"].value
        else:
            raise TypeError("Only ProductType are allowed")

        if isinstance(dictionary["category"], CategoryType):
           dictionary["category"] = dictionary["category"].value
        else:
            raise TypeError("Only CategoryType are allowed")

        return self._request(kwargs.pop('path'), data=Utils.convert_body(dictionary, False), params=kwargs, headers=headers)




from python_paypal_api.base import (
    Client,
    PaypalEndpoint,
    PaypalEndpointParams,
    ApiResponse,
    Utils
)

import logging

class Disputes(Client):


    @PaypalEndpoint('/v1/customer/disputes', method='GET')
    def list_disputes(self, **kwargs) -> ApiResponse: 	
    	return self._request(kwargs.pop('path'), params=kwargs)

    @PaypalEndpoint('/v1/customer/disputes/{}', method='GET')
    def get_dispute(self, disputeId, **kwargs) -> ApiResponse:
        contentType = "application/json"
        headers = {'Content-Type': contentType}
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(disputeId), params=kwargs, headers=headers)

    @PaypalEndpoint('/v1/customer/disputes/{}/accept-offer', method='POST')
    def accept_offer_dispute(self, disputeId, **kwargs) -> ApiResponse:
        return self._request(fPaypalEndpointParams(kwargs.pop('path')).fill(disputeId), params=kwargs, data=kwargs.pop('body'))

    @PaypalEndpoint('/v1/customer/disputes/{}/make-offer', method='POST')
    def make_offer_dispute(self, disputeId, **kwargs) -> ApiResponse:
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(disputeId), params=kwargs, data=kwargs.pop('body'))

    @PaypalEndpoint('/v1/customer/disputes/{}/escalate', method='POST')
    def escalate_dispute(self, disputeId, **kwargs) -> ApiResponse:
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(disputeId), params=kwargs, data=kwargs.pop('body'))

    @PaypalEndpoint('/v1/customer/disputes/{}/send-message', method='POST')
    def send_message(self, disputeId, **kwargs) -> ApiResponse:
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(disputeId), params=kwargs, data=kwargs.pop('body'))

    @PaypalEndpoint('/v1/customer/disputes/{}/provide-evidence', method='POST')
    def provide_evidence(self, disputeId, file, content_type='application/pdf', **kwargs) -> ApiResponse:
        
        fields={
        'input': (None, Utils.convert_body(kwargs.pop('body'), False), 'application/json'),
        'file1': ("sample2.pdf", open(file, 'rb'), 'application/pdf') # could be ignored the application but not the name .pdf
        }

        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(disputeId), params=kwargs, files=fields)

    # Important: This method is for sandbox use only.
    @PaypalEndpoint('/v1/customer/disputes/{}/require-evidence', method='POST')
    def update_dispute(self, disputeId, **kwargs) -> ApiResponse:
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(disputeId), params=kwargs, data=kwargs.pop('body'))
    
    # Important: This method is for sandbox use only.
    @PaypalEndpoint('/v1/customer/disputes/{}/adjudicate', method='POST')
    def adjudicate(self, disputeId, **kwargs) -> ApiResponse:
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(disputeId), params=kwargs, data=kwargs.pop('body'))



from io import BytesIO
from typing import TypeVar
import functools
import hashlib
import base64

class PaypalEndpoint(object):

    def __init__(self, path, method="GET"):
        self.path = path
        self.method = method

    def __call__(self, fn, *args, **kwargs):
        @functools.wraps(fn)
        def decorated(*args, **kwargs):
            kwargs.update({"path": self.path, "method": self.method})
            result = fn(*args, **kwargs)
            return result
        return decorated



class PaypalEndpointParams(object):
     def __init__(self, query: str):
        self.query = query
     def fill(self, *args):
        return self.query.format(*args)



from python_paypal_api.base import (
    Client,
    PaypalEndpoint,
    PaypalEndpointParams,
    ApiResponse,
    Utils
)


class Invoices(Client):


    @PaypalEndpoint('/v2/invoicing/generate-next-invoice-number', method='POST')
    def post_generate_next_invoice_number(self, **kwargs) -> ApiResponse: 
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        return self._request(kwargs.pop('path'), params=kwargs, headers=headers)


    @PaypalEndpoint('/v2/invoicing/invoices', method='GET')
    def get_list_invoices(self, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        return self._request(kwargs.pop('path'), params=kwargs, headers=headers)


    @PaypalEndpoint('/v2/invoicing/invoices', method='POST')
    def post_draft_invoice(self, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        return self._request(kwargs.pop('path'), data=Utils.convert_body(kwargs.pop('body'), False), params=kwargs, headers=headers)


    @PaypalEndpoint('/v2/invoicing/invoices/{}', method='DELETE')
    def post_delete_invoice(self, invoiceId:str, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        Utils.payPalRequestId(headers)
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(invoiceId), params=kwargs, headers=headers)


    @PaypalEndpoint('/v2/invoicing/invoices/{}', method='PUT')
    def put_update_invoice(self, invoiceId:str, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(invoiceId), data=Utils.convert_body(kwargs.pop('body'), False), params=kwargs, headers=headers)


    @PaypalEndpoint('/v2/invoicing/invoices/{}', method='GET')
    def get_invoice(self, invoiceId:str, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(invoiceId), params=kwargs, headers=headers)


    @PaypalEndpoint('/v2/invoicing/invoices/{}/cancel', method='POST')
    def post_cancel_invoice(self, invoiceId:str, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        Utils.payPalRequestId(headers)
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(invoiceId), data=Utils.convert_body(kwargs.pop('body'), False), params=kwargs, headers=headers)


    @PaypalEndpoint('/v2/invoicing/invoices/{}/generate-qr-code', method='POST')
    def post_generate_qr_code(self, invoiceId:str, **kwargs) -> ApiResponse: 
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(invoiceId), data=Utils.convert_body(kwargs.pop('body'), False), params=kwargs, headers=headers)


    @PaypalEndpoint('/v2/invoicing/invoices/{}/payments', method='POST')
    def post_record_payment_invoice(self, invoiceId:str, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(invoiceId), data=Utils.convert_body(kwargs.pop('body'), False), params=kwargs, headers=headers)


    @PaypalEndpoint('/v2/invoicing/invoices/{}/payments/{}', method='DELETE')
    def delete_external_payment_invoice(self, invoiceId:str, transactionId:str, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(invoiceId, transactionId), params=kwargs, headers=headers)


    @PaypalEndpoint('/v2/invoicing/invoices/{}/refunds', method='POST')
    def post_record_refund_invoice(self, invoiceId:str, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(invoiceId), data=Utils.convert_body(kwargs.pop('body'), False), params=kwargs, headers=headers)


    @PaypalEndpoint('/v2/invoicing/invoices/{}/refunds/{}', method='DELETE')
    def delete_external_refund_invoice(self, invoiceId:str, refundId:str, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(invoiceId, refundId), params=kwargs, headers=headers)


    @PaypalEndpoint('/v2/invoicing/invoices/{}/remind', method='POST')
    def post_invoice_remind(self, invoiceId:str, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(invoiceId), data=Utils.convert_body(kwargs.pop('body'), False), params=kwargs, headers=headers)


    @PaypalEndpoint('/v2/invoicing/invoices/{}/send', method='POST')
    def post_send_invoice(self, invoiceId:str, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        Utils.payPalRequestId(headers)
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(invoiceId), data=Utils.convert_body(kwargs.pop('body'), False), params=kwargs, headers=headers)


    @PaypalEndpoint('/v2/invoicing/search-invoices', method='POST')
    def post_search_invoices(self, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        return self._request(kwargs.pop('path'), data=Utils.convert_body(kwargs.pop('body'), False), params=kwargs, headers=headers)


    @PaypalEndpoint('/v2/invoicing/templates', method='GET')
    def get_list_templates(self, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        return self._request(kwargs.pop('path'), params=kwargs, headers=headers)

    @PaypalEndpoint('/v2/invoicing/templates', method='POST')
    def post_create_template(self, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        return self._request(kwargs.pop('path'), data=Utils.convert_body(kwargs.pop('body'), False), params=kwargs, headers=headers)


    @PaypalEndpoint('/v2/invoicing/templates/{}', method='DELETE')
    def delete_template(self, templateId:str, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(templateId), params=kwargs, headers=headers)


    @PaypalEndpoint('/v2/invoicing/templates/{}', method='PUT')
    def put_update_template(self, templateId:str, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(templateId), data=Utils.convert_body(kwargs.pop('body'), False), params=kwargs, headers=headers)


    @PaypalEndpoint('/v2/invoicing/templates/{}', method='GET')
    def get_template(self, templateId:str, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(templateId), params=kwargs, headers=headers)



class BaseConfig:
    def __init__(self, client_id, client_secret):
        self.client_id = client_id
        self.client_secret = client_secret

    def check_config(self):
        errors = []
        for k, v in self.__dict__.items():
            if not v and k != 'refresh_token':
                errors.append(k)
        return errors


class Config(BaseConfig):
    def __init__(self, client_id, client_secret):
        super().__init__(client_id, client_secret)



from python_paypal_api.base import (
    Client,
    PaypalEndpoint,
    PaypalEndpointParams,
    ApiResponse,
    Utils
)

class Identity(Client):

    @PaypalEndpoint('/v1/identity/oauth2/userinfo', method='GET')
    def get_userinfo(self, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json" # headers = {'Content-Type': contentType}
        Utils.contentType(headers, contentType)
        schema = "paypalv1.1"
        kwargs["schema"] = schema
        return self._request(kwargs.pop('path'), params=kwargs, headers=headers)  


    @PaypalEndpoint('/v1/identity/account-settings', method='POST')
    def post_set_account_properties(self, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json" # headers = {'Content-Type': contentType}
        Utils.contentType(headers, contentType)
        return self._request(kwargs.pop('path'), data=Utils.convert_body(kwargs.pop('body'), False), params=kwargs, headers=headers)

    @PaypalEndpoint('/v1/identity/account-settings/deactivate', method='POST')
    def post_disable_account_properties(self, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json" # headers = {'Content-Type': contentType}
        Utils.contentType(headers, contentType)
        return self._request(kwargs.pop('path'), data=Utils.convert_body(kwargs.pop('body'), False), params=kwargs, headers=headers)


import json
import logging
from requests import request
from requests.exceptions import HTTPError
from python_paypal_api.auth.credentials import Credentials
from python_paypal_api.auth import AccessTokenClient, AccessTokenResponse
from python_paypal_api.base.credential_provider import CredentialProvider
from .api_response import ApiResponse
from .base_client import BaseClient
from python_paypal_api.base.enum import EndPoint
from .exceptions import get_exception_for_content, get_exception_for_code
import os
import requests
from io import BytesIO
import gzip
from zipfile import ZipFile
import zipfile
from urllib.parse import urlparse, quote

log = logging.getLogger(__name__)

class Client(BaseClient):

    access_token_client_class = AccessTokenClient
    grantless_scope = ''

    def __init__(
            self,
            account='default',
            credentials=None,
            proxies=None,
            verify=True,
            timeout=None,
            debug=False,
            credential_providers=None
    ):

        super().__init__(account, credentials)

        self.credentials = CredentialProvider(
            account,
            credentials,
            credential_providers=credential_providers,
        ).credentials

        self.host = EndPoint[self.credentials.client_mode].value if self.credentials.client_mode is not None else EndPoint["SANDBOX"].value
        self.endpoint = self.scheme + self.host
        self.debug = debug
        self._auth = self.access_token_client_class(
            account=account,
            credentials=self.credentials,
            proxies=proxies,
            verify=verify,
            timeout=timeout,
        )
        self.timeout = timeout
        self.proxies = proxies
        self.verify = verify

    @property
    def headers(self):
        return {
            'User-Agent': self.user_agent,
            'Authorization': 'Bearer %s' % self.auth.access_token,
            # 'Content-Type': 'application/json'
        }

    @property
    def auth(self) -> AccessTokenResponse:
        return self._auth.get_auth()

    def _request(self,
                 path: str,
                 data: str = None,
                 files = None,
                 params: dict = None,
                 headers = None,
                 ) -> ApiResponse:

        if params is None:
            params = {}

        method = params.pop('method')

        if headers is False:
            base_header = self.headers.copy()
            base_header.pop("Content-Type")
            headers = base_header

        elif headers is not None:

            base_header = self.headers.copy()
            base_header.update(headers)
            headers = base_header

        request_data = data if method in ('POST', 'PUT', 'PATCH') else None

        res = request(
            method,
            self.endpoint + path,
            params=params,
            files=files,
            data=request_data,
            headers=headers or self.headers,
            timeout=self.timeout,
            proxies=self.proxies,
            verify=self.verify,
            )

        if self.debug:
            logging.info(res.request.headers)

            if params:
                message = method + " " + res.request.url
            else:
                message = method + " " + self.endpoint + path

            logging.info(message)
            if data is not None:
                logging.info(data)
            if files is not None:
                logging.info(files)

        return self._check_response(res)

    # @staticmethod
    def _check_response(self, res) -> ApiResponse:

        if self.debug:
            logging.info(vars(res))

        content = vars(res).get('_content')
        headers = vars(res).get('headers')
        status_code = vars(res).get('status_code')


        if status_code == 204:
            data = None
        else:
            str_content = content.decode('utf8')
            data = json.loads(str_content)

        if status_code == 400:
            dictionary = {"name": data["name"], "status_code": vars(res).get('status_code'), "message": data["message"], "details": data["details"]}
            exception = get_exception_for_code(vars(res).get('status_code'))
            raise exception(dictionary, headers=vars(res).get('headers'))

        if status_code == 401:
            dictionary = {"error": data["error"], "status_code": vars(res).get('status_code')}

            try :

                dictionary["error_description"]: data["error_description"]

            except:

                dictionary["error_description"]: data["name"]

            exception = get_exception_for_code(vars(res).get('status_code'))
            raise exception(dictionary, headers=vars(res).get('headers'))

        if status_code == 422:
            # UNPROCESSABLE_ENTITY (The requested action could not be performed, semantically incorrect, or failed business validation.)
            dictionary = {"name": data["name"], "status_code": vars(res).get('status_code'), "details": data["details"]}
            exception = get_exception_for_code(vars(res).get('status_code'))
            raise exception(dictionary, headers=vars(res).get('headers'))

        if status_code == 404:
            # RESOURCE_NOT_FOUND (The specified resource does not exist.)
            dictionary = {"name": data["name"], "status_code": vars(res).get('status_code'), "details": data["details"]}
            exception = get_exception_for_code(vars(res).get('status_code'))
            raise exception(dictionary, headers=vars(res).get('headers'))

        if status_code == 403:
            # RESOURCE_NOT_FOUND (The specified resource does not exist.)
            dictionary = {"name": data["name"], "status_code": vars(res).get('status_code'), "details": data["details"]}
            exception = get_exception_for_code(vars(res).get('status_code'))
            raise exception(dictionary, headers=vars(res).get('headers'))

        next_token = vars(res).get('_next')
        return ApiResponse(data, next_token, headers=headers)



import requests
import hashlib
import logging
import confuse
from python_paypal_api.base import BaseClient
from python_paypal_api.base.enum import EndPoint
from .credentials import Credentials
from .access_token_response import AccessTokenResponse
from .exceptions import AuthorizationError
import os
import logging
import json

from datetime import datetime, timedelta

from requests.auth import HTTPBasicAuth


logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s:%(levelname)s:%(message)s"
)


class AccessTokenClient(BaseClient):

    grant_type = 'client_credentials'
    path = '/v1/oauth2/token'

    def __init__(self, account='default', credentials=None, proxies=None, verify=True, timeout=None):

        self.cred = Credentials(credentials)
        self.host = EndPoint[self.cred.client_mode].value if self.cred.client_mode is not None else EndPoint["SANDBOX"].value
        self.timeout = timeout
        self.proxies = proxies
        self.verify = verify

    def _request(self, url, data, headers):

        response = requests.post(
            url,
            data=data,
            headers=headers,
            timeout=self.timeout,
            proxies=self.proxies,
            verify=self.verify,
            auth=HTTPBasicAuth(data["client_id"], data["client_secret"])
        )

        response_data = response.json()

        if response.status_code != 200:
            error_message = response_data.get('error_description')
            error_code = response_data.get('error')
            raise AuthorizationError(error_code, error_message, response.status_code)
        return response_data


    def create_cache_token(self, file:str):

        now_datetime = datetime.now()
        request_url = self.scheme + self.host + self.path

        try:

            access_token = self._request(request_url, self.data, self.headers)

        except AuthorizationError as error:
            logging.fatal(error)
            exit(0)
        
        future_datetime = now_datetime + timedelta(seconds=access_token["expires_in"])
        access_token["expire_time"] = future_datetime.isoformat()
        json_object = json.dumps(access_token, indent=4)
        flags = os.O_WRONLY | os.O_CREAT
        with os.fdopen(os.open(file, flags, 0o600), 'w') as fout:
            fout.write(json_object)
        return access_token


    def get_auth(self) -> AccessTokenResponse:

        now_datetime = datetime.now()

        config = confuse.Configuration('python-paypal-api')
        file = os.path.join(config.config_dir(), self._get_cache_key())
        try:
            
            openfile = open(file, 'r')
            access_token = json.load(openfile)
            future_datetime = datetime.fromisoformat(access_token["expire_time"])
            # openfile.close()

        except FileNotFoundError:

            access_token = self.create_cache_token(file)
            future_datetime = now_datetime + timedelta(seconds=access_token["expires_in"])
 
        if now_datetime > future_datetime:
            if(os.path.isfile(file)):
                os.remove(file)
            access_token = self.create_cache_token(file)

        else:
            pass

        return AccessTokenResponse(**access_token)


    def authorize_auth_code(self, auth_code):
        request_url = self.scheme + self.host + self.path
        res = self._request(
            request_url,
            data=self._auth_code_request_body(auth_code),
            headers=self.headers
        )
        return res

    def _auth_code_request_body(self, auth_code):
        return {
            'grant_type': 'client_credentials',
            'client_id': self.cred.client_id,
            'client_secret': self.cred.client_secret
        }

    @property
    def data(self):
        return {
            'grant_type': self.grant_type,
            'client_id': self.cred.client_id,
            'client_secret': self.cred.client_secret
        }

    @property
    def headers(self):
        return {
            'User-Agent': self.user_agent,
            'content-type': self.content_type
        }

    def _get_cache_key(self, token_flavor=''):

        return '.' + hashlib.md5(
            (token_flavor + self.cred.client_id).encode('utf-8')
        ).hexdigest()


from python_paypal_api.base import (
    Client,
    PaypalEndpoint,
    PaypalEndpointParams,
    ApiResponse,
    Utils
)

class Transactions(Client):

    @PaypalEndpoint('/v1/reporting/transactions', method='GET')
    def get_list_transactions(self, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        return self._request(kwargs.pop('path'), params=kwargs, headers=headers)


    @PaypalEndpoint('/v1/reporting/balances', method='GET')
    def get_balances(self, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        return self._request(kwargs.pop('path'), params=kwargs, headers=headers)


class PaypalApiException(Exception):
    code = 999

    def __init__(self, error, headers):

        try:
            self.message = error.get('message')
            self.details = error.get('details')
            self.paypal_code = error.get('status_code')
        except IndexError:
            pass

        self.error = error
        self.headers = headers


class PaypalTypeException(TypeError):
    def __init__(self, error):
        super(TypeError, self).__init__(error)

class PaypalApiRequestException(PaypalApiException):
    code = 400

    def __init__(self, error, headers):

        try:
            self.name = error.get('name')
            self.message = error.get('message')
            self.details = error.get('details')
            self.paypal_code = error.get('status_code')
        except IndexError:
            pass

        self.error = error
        self.headers = headers


class PaypalApiBadRequestException(PaypalApiRequestException):
    """
    400 Request has missing or invalid parameters and cannot be parsed.
    """
    def __init__(self, error, headers=None):
        super(PaypalApiBadRequestException, self).__init__(error, headers)

class PaypalApiUnprocessableEntityException(PaypalApiRequestException):
    """
    400 Request has missing or invalid parameters and cannot be parsed.
    """
    def __init__(self, error, headers=None):
        super(PaypalApiUnprocessableEntityException, self).__init__(error, headers)


class PaypalApiForbiddenException(PaypalApiException):
    """
    401 Indicates access to the resource is forbidden. Possible reasons include Access Denied, Unauthorized, Expired Token, or Invalid Signature.
    """

    def __init__(self, error, headers=None):
        super(PaypalApiForbiddenException, self).__init__(error, headers)


class PaypalApiResourceNotFound(PaypalApiException):
    """
    401 Indicates access to the resource is forbidden. Possible reasons include Access Denied, Unauthorized, Expired Token, or Invalid Signature.
    """
    def __init__(self, error, headers=None):
        super(PaypalApiResourceNotFound, self).__init__(error, headers)


def get_exception_for_code(code: int):
    return {
        400: PaypalApiBadRequestException,
        422: PaypalApiUnprocessableEntityException,
        401: PaypalApiForbiddenException,
        403: PaypalApiForbiddenException,
        404: PaypalApiResourceNotFound
    }.get(code, PaypalApiException)




def get_exception_for_content(content: object):
    return {
        'UNAUTHORIZED': PaypalApiForbiddenException
    }.get(content.get('code'), PaypalApiException)



from python_paypal_api.base import (
    Client,
    PaypalEndpoint,
    PaypalEndpointParams,
    ApiResponse,
    Utils
)

import logging

class Tracking(Client):


    @PaypalEndpoint('/v1/shipping/trackers/{}', method='PUT')
    def put_traking(self, id: str, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json" # headers = {'Content-Type': contentType}
        Utils.contentType(headers, contentType)
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(id), data=Utils.convert_body(kwargs.pop('body'), False), params=kwargs, headers=headers)


    @PaypalEndpoint('/v1/shipping/trackers-batch', method='POST')
    def post_traking(self, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json" # headers = {'Content-Type': contentType}
        Utils.contentType(headers, contentType)
        return self._request(kwargs.pop('path'), data=Utils.convert_body(kwargs.pop('body'), False), params=kwargs, headers=headers)

    @PaypalEndpoint('/v1/shipping/trackers/{}', method='GET')
    def get_traking(self, id: str, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json" # headers = {'Content-Type': contentType}
        Utils.contentType(headers, contentType)
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(id), params=kwargs, headers=headers)    




import python_paypal_api.version as vd
from enum import Enum

class BaseClient:
    scheme = 'https://'
    method = 'GET'
    content_type = 'application/x-www-form-urlencoded;charset=UTF-8'
    user_agent = 'python-paypal-api'

    def __init__(self, account='default', credentials=None):

        try:
            version = vd.__version__
            self.user_agent += f'-{version}'
        except:
            pass


from python_paypal_api.base import (
    Client,
    PaypalEndpoint,
    PaypalEndpointParams,
    ApiResponse,
    Utils
)


class PartnerReferrals(Client):

    @PaypalEndpoint('/v2/customer/partner-referrals', method='POST')
    def post_create_partner_referrals(self, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        return self._request(kwargs.pop('path'), data=Utils.convert_body(kwargs.pop('body'), False), params=kwargs, headers=headers)


    @PaypalEndpoint('/v2/customer/partner-referrals/{}', method='GET')
    def get_referral_data(self, partnerReferralId:str, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(partnerReferralId), params=kwargs, headers=headers)


import abc
import functools
import json
import os

from typing import Dict, Iterable, Optional, Type

import confuse
import logging
# from cachetools import Cache

logger = logging.getLogger(__name__)
required_credentials = [
    'client_id',
    'client_secret'
]


class MissingCredentials(Exception):
    """
    Credentials are missing, see the error output to find possible causes
    """
    pass


class BaseCredentialProvider(abc.ABC):
    errors = []
    credentials = None

    def __init__(self, account: str = 'default', *args, **kwargs):
        self.account = account

    def __call__(self, *args, **kwargs):
        self.load_credentials()
        return self.check_credentials()

    @abc.abstractmethod
    def load_credentials(self):
        ...

    def check_credentials(self):
        try:
            self.errors = [c for c in required_credentials if
                           c not in self.credentials.keys() or not self.credentials[c]]
        except (AttributeError, TypeError):
            raise MissingCredentials(f'Credentials are missing: {", ".join(required_credentials)}')
        if not len(self.errors):
            return self.credentials
        raise MissingCredentials(f'Credentials are missing: {", ".join(self.errors)}')


class FromCodeCredentialProvider(BaseCredentialProvider):

    def load_credentials(self):
        return None

    def __init__(self, credentials: dict, *args, **kwargs):
        super(FromCodeCredentialProvider, self).__init__('default', credentials)
        self.credentials = credentials

class FromConfigFileCredentialProvider(BaseCredentialProvider):

    def load_credentials(self):

        try:
            config = confuse.Configuration('python-paypal-api')
            config_filename = os.path.join(config.config_dir(), 'credentials.yml')
            config.set_file(config_filename)
            account_data = config[self.account].get()
            self.credentials = account_data

        except (confuse.exceptions.NotFoundError, confuse.exceptions.ConfigReadError):
            logging.info(confuse.exceptions.NotFoundError)
            return


class CredentialProvider:
    credentials = None
    logger.info("CredentialProvider")

    CREDENTIAL_PROVIDERS: Iterable[Type[BaseCredentialProvider]] = (
        FromCodeCredentialProvider,
        FromConfigFileCredentialProvider
    )

    def __init__(
        self,
        account: str = 'default',
        credentials: Optional[Dict[str, str]] = None,
        credential_providers: Optional[Iterable[Type[BaseCredentialProvider]]] = None,
    ):
        self.account = account
        providers = self.CREDENTIAL_PROVIDERS if credential_providers is None else credential_providers
        for cp in providers:
            try:
                self.credentials = cp(account=account, credentials=credentials)()
                break
            except MissingCredentials:
                continue
        if self.credentials:
            self.credentials = self.Config(**self.credentials)
        else:
            raise MissingCredentials(f'Credentials are missing: {", ".join(required_credentials)}')

    class Config:
        def __init__(self, **kwargs):
            self.client_id = kwargs.get('client_id')
            self.client_secret = kwargs.get('client_secret')
            self.client_mode = kwargs.get('client_mode')




from .access_token_client import AccessTokenClient
from .access_token_response import AccessTokenResponse
from .credentials import Credentials

__all__ = [
    'AccessTokenResponse',
    'AccessTokenClient',
    'Credentials',
]



from enum import Enum

class EndPoint(Enum):
    PRODUCTION = "api.paypal.com"
    SANDBOX = "api.sandbox.paypal.com"

class ProductType(str, Enum):
    PHYSICAL = 'PHYSICAL'
    DIGITAL = 'DIGITAL'
    SERVICE = 'SERVICE'

class CategoryType(str, Enum):
    AC_REFRIGERATION_REPAIR = 'AC_REFRIGERATION_REPAIR'
    ACADEMIC_SOFTWARE = 'ACADEMIC_SOFTWARE'
    ACCESSORIES = 'ACCESSORIES'
    # Check how to find the whole categories in code mode


class AccessTokenResponse:
    def __init__(self, **kwargs):
        self.scope = kwargs.get('scope')
        self.access_token = kwargs.get('access_token')
        self.token_type = kwargs.get('token_type')
        self.app_id = kwargs.get('app_id')
        self.expires_in = kwargs.get('expires_in')
        self.nonce = kwargs.get('nonce')




class AuthorizationError(Exception):
    """
    Authorization Error

    Parameters:

        error_code: str Error code from paypal auth api
        error_msg: str Error sm
        status_code: integer Response status code from paypal auth api
    """

    def __init__(self, error_code, error_msg, status_code):
        self.error_code = error_code
        self.message = error_msg
        self.status_code = status_code



from .orders import Orders
from .disputes import Disputes
from .catalog import Catalog
from .tracking import Tracking
from .identity import Identity
from .invoices import Invoices
from .partner_referrals import PartnerReferrals
from .transactions import Transactions

__all__ = [
    "Orders",
    "Disputes",
    "Catalog", 
    "Identity",
    "Invoices",
    "PartnerReferrals",
    "Transactions"
]


import time
import os
import json
from json.decoder import JSONDecodeError
from io import TextIOWrapper
from python_paypal_api.base import PaypalTypeException
import uuid

class Utils:

    @staticmethod
    def payPalRequestId(headers: dict):
        headers["PayPal-Request-Id"] = str(uuid.uuid1())
        return headers

    def payPalClientMetadataId(headers: dict):
        headers["PayPal-Client-Metadata-Id"] = "p5AngF3dfz5uYeQq4Tjws21Qx48Z0Nt9cCX3"

    @staticmethod
    def prefer(headers: dict, value: str):
        headers["Prefer"] = value
        return headers
        
    @staticmethod
    def contentType(headers: dict, value: str):
        headers["Content-Type"] = value
        return headers

    @staticmethod
    def clean_dictionary(dictionary: dict, valid_keys: set):
        """Cleans a dictionary removing null entries and invalid keys
        
        Arguments:
            dictionary {dict} -- The dictionary to be checked
            valid_keys {set} -- The keys that should not be deleted
        """
        keys = list(dictionary.keys())
        for k in keys:
            if dictionary.get(k) == None or not k in valid_keys:
                del dictionary[k] 
        return dictionary


    @staticmethod
    def convert_updates(body):

        comp_keys = ['op', 'path', 'value']

        if isinstance(body, list):

            if isinstance(body[0], tuple):
                body = [ {'op': x[0], 'path': x[1], 'value': x[2]} for x in body ]

            elif isinstance(body[0], dict):
                res = all(sorted(list(body[x].keys())) == sorted(comp_keys) for x in range(len(body)))

        elif isinstance(body, dict):
            body = [body]

        elif isinstance(body, str):
            obj = json.loads(body)

            if isinstance(obj, dict):
                body = [obj]

            elif isinstance(obj, list):
                return body
        else:
            error = "Only type list, dict or str are allowed on body"
            raise PaypalTypeException('TypeError: '+ error)

        return(json.dumps(body))


    @staticmethod
    def convert_body(body, wrap: bool = True):

        if isinstance(body, str):

            if os.path.isfile(body):
                body = open(body, mode="r", encoding="utf-8")
                body = body.read()
                try:
                    json.loads(body)
                except JSONDecodeError as error:
                    raise AdvertisingTypeException(f"{type(error)}", error)
            else:
                try:
                    body = json.loads(body)
                except ValueError as error:
                    raise AdvertisingTypeException(f"{type(error)}", error)
                pass

        if isinstance(body, dict) and wrap:
            try:
                body = json.dumps([body])
            except TypeError as error:
                raise AdvertisingTypeException(f"{type(error)}", error)

        if isinstance(body, dict) and wrap is False:
            try:
                body = json.dumps(body)
            except TypeError as error:
                raise AdvertisingTypeException(f"{type(error)}", error)

        if isinstance(body, list):
            try:
                body = json.dumps(body)
            except TypeError as error:
                raise AdvertisingTypeException(f"{type(error)}", error)

        if isinstance(body, TextIOWrapper):
            body = body.read()
            try:
                json.loads(body)
            except JSONDecodeError as error:
                raise AdvertisingTypeException(f"{type(error)}", error)

        return body

    def load_all_pages(throttle_by_seconds: float = 2, next_token_param='NextToken',
                       use_rate_limit_header: bool = False,
                       extras: dict = None):
        """
        Load all pages if a next token is returned

        Args:
            throttle_by_seconds: float
            next_token_param: str | The param amazon expects to hold the next token
            use_rate_limit_header: if the function should try to use amazon's rate limit header
            extras: additional data to be sent with NextToken, e.g `dict(QueryType='NEXT_TOKEN')` for `FulfillmentInbound`
        Returns:
            Transforms the function in a generator, returning all pages
        """
        if not extras:
            extras = {}

        def decorator(function):
            def wrapper(*args, **kwargs):
                res = function(*args, **kwargs)
                yield res
                if "nextCursor" in res.payload.get("payload"):
                    kwargs.clear()
                    kwargs.update({next_token_param: res.payload.get("payload")["nextCursor"], **extras})
                    sleep_time = throttle_by_seconds
                    for x in wrapper(*args, **kwargs):
                        yield x
                        if sleep_time > 0:
                            time.sleep(throttle_by_seconds)

            wrapper.__doc__ = function.__doc__
            return wrapper

        return decorator


    def load_all_categories(throttle_by_seconds: float = 2, next_token_param='NextToken',
                       use_rate_limit_header: bool = False,
                       extras: dict = None):
        """
        Load all pages if a next token is returned

        Args:
            throttle_by_seconds: float
            next_token_param: str | The param amazon expects to hold the next token
            use_rate_limit_header: if the function should try to use amazon's rate limit header
            extras: additional data to be sent with NextToken, e.g `dict(QueryType='NEXT_TOKEN')` for `FulfillmentInbound`
        Returns:
            Transforms the function in a generator, returning all pages
        """
        if not extras:
            extras = {}

        def decorator(function):
            def wrapper(*args, **kwargs):
                res = function(*args, **kwargs)
                yield res
                if next_token_param in res.payload:
                    # kwargs.clear()
                    kwargs.update({next_token_param: res.payload[next_token_param], **extras})
                    sleep_time = throttle_by_seconds
                    for x in wrapper(*args, **kwargs):
                        yield x
                        if sleep_time > 0:
                            time.sleep(throttle_by_seconds)

            wrapper.__doc__ = function.__doc__
            return wrapper

        return decorator


from .base_client import BaseClient
from .client import Client
from .helpers import PaypalEndpoint, PaypalEndpointParams
from .exceptions import PaypalApiException, PaypalApiRequestException, PaypalTypeException
from .exceptions import PaypalApiBadRequestException, PaypalApiForbiddenException
from .exceptions import PaypalApiUnprocessableEntityException
from .exceptions import PaypalApiResourceNotFound
from .credential_provider import CredentialProvider, MissingCredentials
from .api_response import ApiResponse
from .utils import Utils
from .enum import ProductType, CategoryType, EndPoint

__all__ = [
    'AccessTokenClient',
    'ApiResponse',
    'Client',
    'BaseClient',
    'PaypalEndpointParams',
    'PaypalEndpoint',
    'PaypalApiException',
    'PaypalApiRequestException',
    'PaypalApiBadRequestException',
    'PaypalApiUnprocessableEntityException',
    'PaypalApiResourceNotFound',
    'PaypalApiForbiddenException'
    'CredentialProvider',
    'MissingCredentials',
    'Utils',
    'ProductType',
    'CategoryType',
    'EndPoint'
]



class Credentials:
    def __init__(self, credentials):
        self.client_id = credentials.client_id
        self.client_secret = credentials.client_secret
        self.client_mode = credentials.client_mode


from python_paypal_api.base import (
    Client,
    PaypalEndpoint,
    PaypalEndpointParams,
    ApiResponse,
    Utils
)

import logging

class Orders(Client):


    @PaypalEndpoint('/v2/checkout/orders', method='POST')
    def post_create_order(self, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json" # headers = {'Content-Type': contentType}
        prefer = "return=representation" # return=minimal return=representation
        Utils.contentType(headers, contentType)
        Utils.payPalRequestId(headers)
        return self._request(kwargs.pop('path'), data=Utils.convert_body(kwargs.pop('body'), False), params=kwargs, headers=headers)


    @PaypalEndpoint('/v2/checkout/orders/{}', method='PATCH')
    def patch_update_order(self, orderId:str, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json"
        Utils.contentType(headers, contentType)
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(orderId), data=Utils.convert_body(kwargs.pop('body'), False), params=kwargs, headers=headers)

    @PaypalEndpoint('/v2/checkout/orders/{}', method='GET')
    def get_order(self, orderId:str, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json" # headers = {'Content-Type': contentType}
        Utils.contentType(headers, contentType)
    	# return self._request(fill_query_params(kwargs.pop('path'), orderId), params=kwargs, headers=headers)
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(orderId), params=kwargs, headers=headers)

    @PaypalEndpoint('/v2/checkout/orders/{}/capture', method='POST')
    def post_capture_for_order(self, orderId:str, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json" # headers = {'Content-Type': contentType}
        prefer = "return=representation" # return=minimal return=representation
        Utils.contentType(headers, contentType)
        Utils.payPalRequestId(headers)
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(orderId), params=kwargs, headers=headers)


    @PaypalEndpoint('/v2/checkout/orders/{}/authorize', method='POST')
    def post_authorize_for_order(self, orderId:str, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json" # headers = {'Content-Type': contentType}
        prefer = "return=representation" # return=minimal return=representation
        Utils.contentType(headers, contentType)
        Utils.payPalRequestId(headers)
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(orderId), params=kwargs, headers=headers)


    @PaypalEndpoint('/v2/checkout/orders/{}/confirm-payment-source', method='POST')
    def post_confirm_order(self, orderId:str, **kwargs) -> ApiResponse:
        headers = {}
        contentType = "application/json" # headers = {'Content-Type': contentType}
        prefer = "return=representation" # return=minimal return=representation
        Utils.contentType(headers, contentType)
        Utils.payPalRequestId(headers)
        return self._request(PaypalEndpointParams(kwargs.pop('path')).fill(orderId), data=Utils.convert_body(kwargs.pop('body'), False), params=kwargs, headers=headers)


